import datetime
import ssl
from asyncio import AbstractEventLoop, gather
from typing import List
from unittest.mock import AsyncMock, patch

import aiodns
import pytest
from aiohttp.http_exceptions import HttpProcessingError
from aioresponses import CallbackResult, aioresponses
from pydantic import BaseModel
from yarl import URL

import biar
import biar.errors

BASE_URL = "https://api.com/v1"


@pytest.fixture
def mock_server() -> aioresponses:
    with aioresponses() as m:
        yield m


class TestRequest:
    @pytest.mark.asyncio
    async def test_request_structured_many(self, mock_server: aioresponses):
        # arrange
        class MyModel(BaseModel):
            key: str

        headers = {"Content-Type": "application/json"}
        response_json_content = {"key": "value"}
        for i in range(2):
            mock_server.get(
                url=URL(BASE_URL),
                headers=headers,
                payload=response_json_content,
            )
        target_response = [
            biar.model.StructuredResponse(
                url=URL(BASE_URL),
                status_code=200,
                headers=headers,
                json_content=response_json_content,
                structured_content=MyModel(key="value"),
            )
        ] * 2

        # act
        output_response = await biar.request_structured_many(
            model=MyModel,
            urls=[BASE_URL] * 2,
            config=biar.RequestConfig(download_text_content=False),
        )

        # assert
        assert target_response == output_response

    @pytest.mark.asyncio
    async def test_request_many_payload(self, mock_server: aioresponses):
        # arrange
        def callback(url: URL, **kwargs):
            if url == URL(BASE_URL) / "1" and kwargs["json"] == {
                "key": "1",
                "ts": "2023-01-01T00:00:00",
            }:
                return CallbackResult(status=200)
            elif url == URL(BASE_URL) / "2" and kwargs["json"] == {
                "key": "2",
                "ts": "2023-01-02T00:00:00",
            }:
                return CallbackResult(status=200)
            return CallbackResult(status=500)

        mock_server.put(url=URL(BASE_URL) / "1", callback=callback)
        mock_server.put(url=URL(BASE_URL) / "2", callback=callback)

        class Payload(BaseModel):
            key: str
            ts: datetime.datetime

        # act
        output_responses = await biar.request_many(
            urls=[URL(BASE_URL) / "1", URL(BASE_URL) / "2"],
            config=biar.RequestConfig(method="PUT"),
            payloads=[
                Payload(key="1", ts="2023-01-01T00:00:00"),
                Payload(key="2", ts="2023-01-02T00:00:00"),
            ],
        )

        # assert
        assert all([response.status_code == 200 for response in output_responses])

    @pytest.mark.asyncio
    async def test_request_retry_success(self, mock_server: aioresponses):
        # arrange
        mock_server.get(url=BASE_URL, status=500)
        mock_server.get(url=BASE_URL, status=200)

        # act
        start_ts = datetime.datetime.utcnow()
        output_response = await biar.request(
            url=BASE_URL,
            config=biar.RequestConfig(
                method="GET",
                download_json_content=False,
                retryer=biar.model.Retryer(
                    attempts=2,
                    min_delay=1,
                    max_delay=1,
                ),
            ),
        )
        end_ts = datetime.datetime.utcnow()
        elapsed_time = (end_ts - start_ts).total_seconds()

        # assert
        assert output_response.status_code == 200
        assert 1 < elapsed_time < 2

    @pytest.mark.asyncio
    async def test_request_retry_exception(self, mock_server: aioresponses):
        # arrange
        mock_server.get(url=BASE_URL, status=500)
        mock_server.get(url=BASE_URL, exception=HttpProcessingError())
        mock_server.get(url=BASE_URL, exception=HttpProcessingError())
        retrier = biar.model.Retryer(attempts=2, max_delay=0)

        # act and assert
        with pytest.raises(HttpProcessingError):
            _ = await biar.request(
                url=BASE_URL,
                config=biar.RequestConfig(
                    method="GET",
                    retryer=retrier,
                ),
            )

    @pytest.mark.asyncio
    async def test_request_retry_status_fail(self, mock_server: aioresponses):
        # arrange
        mock_server.get(url=BASE_URL, status=500)
        mock_server.get(url=BASE_URL, status=500)
        mock_server.get(url=BASE_URL, status=500)
        retrier = biar.model.Retryer(attempts=2, max_delay=0)

        # act and assert
        with pytest.raises(biar.errors.ResponseEvaluationError):
            _ = await biar.request(
                url=BASE_URL,
                config=biar.RequestConfig(
                    method="GET",
                    retryer=retrier,
                ),
            )

    def test_request_rate_limit(
        self, event_loop: AbstractEventLoop, mock_server: aioresponses
    ):
        # arrange
        mock_server.get(url=BASE_URL, status=200)
        mock_server.get(url=BASE_URL, status=200)
        mock_server.get(url=BASE_URL, status=200)
        rate_limiter = biar.model.RateLimiter(rate=2, time_frame=1, identity="api")
        config = biar.RequestConfig(
            method="GET",
            rate_limiter=rate_limiter,
        )
        async_requests = [
            biar.request(url=BASE_URL, config=config),
            biar.request(url=BASE_URL, config=config),
            biar.request(url=BASE_URL, config=config),
        ]

        # act
        start_ts = datetime.datetime.utcnow()
        _ = event_loop.run_until_complete(gather(*async_requests))
        end_ts = datetime.datetime.utcnow()
        elapsed_time = (end_ts - start_ts).total_seconds()

        # assert
        assert 1 < elapsed_time < 2

    @pytest.mark.asyncio
    async def test_request_structured_many_value_error(self):
        # arrange
        class MyModel(BaseModel):
            key: str

        # act and assert
        with pytest.raises(ValueError):
            _ = await biar.request_structured_many(
                model=MyModel,
                urls=[BASE_URL] * 2,
                config=biar.RequestConfig(download_text_content=False),
                payloads=[MyModel(key="value")],
            )

    @pytest.mark.asyncio
    async def test_request_structured_retry_on_content(self, mock_server: aioresponses):
        # arrange
        class MyModel(BaseModel):
            key: str
            items: List[int]

        mock_server.get(url=BASE_URL, status=200, payload={"key": "value", "items": []})
        mock_server.get(url=BASE_URL, status=200, payload={"key": "value", "items": []})
        mock_server.get(
            url=BASE_URL, status=200, payload={"key": "value", "items": [1, 2, 3]}
        )

        # act
        output_response = await biar.request_structured(
            model=MyModel,
            url=BASE_URL,
            config=biar.RequestConfig(
                method="GET",
                retryer=biar.model.Retryer(
                    attempts=3,
                    max_delay=0,
                    retry_based_on_content_callback=lambda x: not x.items,
                ),
            ),
        )

        # assert
        assert output_response.structured_content.items == [1, 2, 3]

    @pytest.mark.asyncio
    async def test_poll(self, mock_server: aioresponses):
        # arrange
        class MyModel(BaseModel):
            status: str

        mock_server.get(url=BASE_URL, status=200, payload={"status": "pending"})
        mock_server.get(url=BASE_URL, status=200, payload={"status": "pending"})
        mock_server.get(url=BASE_URL, status=200, payload={"status": "success"})

        # act
        output_response = await biar.poll(
            model=MyModel,
            poll_config=biar.PollConfig(
                timeout=1, interval=0, success_condition=lambda x: x.status == "success"
            ),
            url=BASE_URL,
        )

        # assert
        assert output_response.structured_content.status == "success"

    @pytest.mark.asyncio
    async def test_poll_timeout(self, mock_server: aioresponses):
        # arrange
        class MyModel(BaseModel):
            status: str

        mock_server.get(url=BASE_URL, status=200, payload={"status": "pending"})
        mock_server.get(url=BASE_URL, status=200, payload={"status": "pending"})
        mock_server.get(url=BASE_URL, status=200, payload={"status": "pending"})

        # act and assert
        with pytest.raises(biar.errors.PollError):
            _ = await biar.poll(
                model=MyModel,
                poll_config=biar.PollConfig(
                    timeout=0.1,
                    interval=0.05,
                    success_condition=lambda x: x.status == "success",
                ),
                url=BASE_URL,
            )


def test_get_ssl_context():
    # arrange
    extra_certificate = """
-----BEGIN CERTIFICATE-----
MIICVDCCAdugAwIBAgIQZ3SdjXfYO2rbIvT/WeK/zjAKBggqhkjOPQQDAzBsMQsw
CQYDVQQGEwJHUjE3MDUGA1UECgwuSGVsbGVuaWMgQWNhZGVtaWMgYW5kIFJlc2Vh
cmNoIEluc3RpdHV0aW9ucyBDQTEkMCIGA1UEAwwbSEFSSUNBIFRMUyBFQ0MgUm9v
dCBDQSAyMDIxMB4XDTIxMDIxOTExMDExMFoXDTQ1MDIxMzExMDEwOVowbDELMAkG
A1UEBhMCR1IxNzA1BgNVBAoMLkhlbGxlbmljIEFjYWRlbWljIGFuZCBSZXNlYXJj
aCBJbnN0aXR1dGlvbnMgQ0ExJDAiBgNVBAMMG0hBUklDQSBUTFMgRUNDIFJvb3Qg
Q0EgMjAyMTB2MBAGByqGSM49AgEGBSuBBAAiA2IABDgI/rGgltJ6rK9JOtDA4MM7
KKrxcm1lAEeIhPyaJmuqS7psBAqIXhfyVYf8MLA04jRYVxqEU+kw2anylnTDUR9Y
STHMmE5gEYd103KUkE+bECUqqHgtvpBBWJAVcqeht6NCMEAwDwYDVR0TAQH/BAUw
AwEB/zAdBgNVHQ4EFgQUyRtTgRL+BNUW0aq8mm+3oJUZbsowDgYDVR0PAQH/BAQD
AgGGMAoGCCqGSM49BAMDA2cAMGQCMBHervjcToiwqfAircJRQO9gcS3ujwLEXQNw
SaSS6sUUiHCm0w2wqsosQJz76YJumgIwK0eaB8bRwoF8yguWGEEbo/QwCZ61IygN
nxS2PFOiTAZpffpskcYqSUXm7LcT4Tps
-----END CERTIFICATE-----
"""
    # act
    ssl_context = biar.get_ssl_context(extra_certificate=extra_certificate)

    # assert
    isinstance(ssl_context, ssl.SSLContext)


class TestIsHostReachableService:
    @pytest.mark.asyncio
    async def test_is_host_reachable(self):
        # act
        with patch("aiodns.DNSResolver.query", new=AsyncMock()) as mock:
            output = await biar.is_host_reachable(host="google.com")

        # assert
        assert output is True
        mock.assert_called_once_with("google.com", qtype="A")

    @pytest.mark.asyncio
    async def test_is_host_reachable_error(self):
        # arrange
        async def side_effect(*args, **kwargs):
            raise aiodns.error.DNSError

        # act
        with patch("aiodns.DNSResolver.query", side_effect=side_effect) as mock:
            output = await biar.is_host_reachable(host="google.com")

        # assert
        assert output is False
        assert "google.com" in mock.mock_calls[0].args



"""Metadata module."""

__title__: str = "biar"
__description__: str = "batteries-included async requests tool for python."
__version__: str = "0.5.1"
__url__: str = "https://github.com/rafaelleinio/biar"
__author__: str = "Rafael Leinio"
__license__: str = "MIT"



class ResponseEvaluationError(Exception):
    """Base Exception for non-OK responses."""


class PollError(Exception):
    """Base Exception for poll errors."""


class ContentCallbackError(Exception):
    """Base Exception for content callback errors."""



from biar.__metadata__ import (
    __author__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__,
)
from biar.errors import ContentCallbackError, PollError, ResponseEvaluationError
from biar.model import (
    PollConfig,
    ProxyConfig,
    RateLimiter,
    RequestConfig,
    Response,
    Retryer,
    StructuredResponse,
)
from biar.services import (
    get_ssl_context,
    is_host_reachable,
    poll,
    request,
    request_many,
    request_structured,
    request_structured_many,
)

__all__ = [
    "__title__",
    "__description__",
    "__version__",
    "__url__",
    "__author__",
    "__license__",
    "ProxyConfig",
    "RateLimiter",
    "Response",
    "RequestConfig",
    "Retryer",
    "StructuredResponse",
    "get_ssl_context",
    "is_host_reachable",
    "request",
    "request_structured",
    "request_structured_many",
    "ResponseEvaluationError",
    "request_many",
    "poll",
    "PollConfig",
    "PollError",
    "ContentCallbackError",
]



import asyncio
import datetime
import ssl
from typing import Any, Callable, Dict, List, Optional, Type, Union

import aiodns
import aiohttp
import certifi
import tenacity
from loguru import logger
from pydantic import BaseModel
from yarl import URL

from biar import (
    ContentCallbackError,
    PollConfig,
    PollError,
    RateLimiter,
    RequestConfig,
    Response,
    ResponseEvaluationError,
    StructuredResponse,
)
from biar.user_agents import get_user_agent


async def is_host_reachable(host: str) -> bool:
    """Async check if a host is reachable.

    Args:
        host: url to check if is reachable.

    Returns:
        True if the host is reachable.

    """
    dns_solver = aiodns.DNSResolver()
    try:
        _ = await dns_solver.query(host, qtype="A")
        return True
    except aiodns.error.DNSError:
        return False


def get_ssl_context(extra_certificate: Optional[str] = None) -> ssl.SSLContext:
    """Create a ssl context.

    It uses the collection of certificates provided by certifi package. Besides, the
    user can give an additional certificate to be appended to the final collection.

    Args:
        extra_certificate: extra string certificate to be used alongside default ones.

    Returns:
        new ssl context.

    """
    with open(certifi.where()) as f:
        certificate = f.read()
    if extra_certificate:
        certificate = certificate + "\n" + extra_certificate
    return ssl.create_default_context(cadata=certificate)


async def _request_base(
    download_json_content: bool,
    download_text_content: bool,
    rate_limiter: RateLimiter,
    session: aiohttp.ClientSession,
    acceptable_codes: Optional[List[int]] = None,
    **request_kwargs: Any,
) -> Response:
    rate_limiter.limiter.try_acquire(name=rate_limiter.identity)
    async with session.request(**request_kwargs) as response:
        text_content = await response.text() if download_text_content else ""
        if response.status not in (acceptable_codes or [200]):
            raise ResponseEvaluationError(
                f"Error: status={response.status}, "
                f"Text content (if loaded): {text_content}"
            )

        json_content = await response.json() if download_json_content else None
        normalized_json_content = (
            json_content
            if isinstance(json_content, dict)
            else {"content": json_content}
        )
        http_response = Response(
            url=response.url,
            status_code=response.status,
            headers={k: v for k, v in response.headers.items()},
            json_content=normalized_json_content,
            text_content=text_content,
        )

    return http_response


@tenacity.retry
async def _request(
    download_json_content: bool,
    download_text_content: bool,
    rate_limiter: RateLimiter,
    session: aiohttp.ClientSession,
    acceptable_codes: Optional[List[int]] = None,
    **request_kwargs: Any,
) -> Response:
    return await _request_base(
        download_json_content=download_json_content,
        download_text_content=download_text_content,
        rate_limiter=rate_limiter,
        session=session,
        acceptable_codes=acceptable_codes,
        **request_kwargs,
    )


def _build_kwargs(
    url: Union[str, URL],
    config: RequestConfig,
    payload: Optional[BaseModel] = None,
) -> Dict[str, Any]:
    headers = {
        **(config.headers or {}),
        **(
            {"User-Agent": get_user_agent(user_agent_list=config.user_agent_list)}
            if config.use_random_user_agent
            else {}
        ),
        **(
            {"Authorization": f"Bearer {config.bearer_token}"}
            if config.bearer_token
            else {}
        ),
    }
    proxy_kwargs = (
        {
            "proxy": config.proxy_config.host,
            "proxy_headers": config.proxy_config.headers,
            "ssl_context": get_ssl_context(
                extra_certificate=config.proxy_config.ssl_cadata
            ),
        }
        if config.proxy_config
        else {}
    )
    return {
        "url": url,
        "method": config.method,
        "headers": headers,
        "params": config.params or None,
        "timeout": config.timeout,
        "json": payload.model_dump(mode="json") if payload else None,
        **proxy_kwargs,
    }


async def request(
    url: Union[str, URL],
    config: RequestConfig = RequestConfig(),
    payload: Optional[BaseModel] = None,
) -> Response:
    """Make a request.

    Args:
        url: url to send request.
        config: request configuration.
        payload: payload to be sent in the request as a structured pydantic model.

    Returns:
        Response object from the request.

    """
    logger.debug(f"Request started, {config.method} method to {url}...")
    new_callable = _request.retry_with(**config.retryer.retrying_config)  # type: ignore

    async with aiohttp.ClientSession() as new_session:
        response: Response = await new_callable(
            download_json_content=config.download_json_content,
            download_text_content=config.download_text_content,
            rate_limiter=config.rate_limiter,
            session=config.session or new_session,
            acceptable_codes=config.acceptable_codes,
            **_build_kwargs(url=url, config=config, payload=payload),
        )

    logger.debug("Request finished!")
    return response


def _normalize_payloads(
    urls: List[Union[str, URL]],
    payloads: Optional[List[BaseModel]] = None,
) -> Optional[List[BaseModel]]:
    payloads = payloads or []
    if payloads and len(urls) != len(payloads):
        raise ValueError(
            f"Number of urls ({len(urls)}) and payloads ({len(payloads or [])}) "
            f"must be the same."
        )
    return payloads


async def request_many(
    urls: List[Union[str, URL]],
    config: RequestConfig = RequestConfig(),
    payloads: Optional[List[BaseModel]] = None,
) -> List[Response]:
    """Make many requests.

    Args:
        urls: list of urls to send requests.
        config: request configuration.
        payloads: list of payloads as structured pydantic models.

    Returns:
        List of response objects from the requests.

    """
    payloads = _normalize_payloads(urls=urls, payloads=payloads)
    coroutines = (
        [
            request(
                url=url,
                config=config,
                payload=payload,
            )
            for url, payload in zip(urls, payloads)
        ]
        if payloads
        else [
            request(
                url=url,
                config=config,
            )
            for url in urls
        ]
    )

    results: List[Response] = await asyncio.gather(*coroutines)
    return results


@tenacity.retry
async def _request_structured(
    model: Type[BaseModel],
    retry_based_on_content_callback: Optional[Callable[[StructuredResponse], bool]],
    download_json_content: bool,
    download_text_content: bool,
    rate_limiter: RateLimiter,
    session: aiohttp.ClientSession,
    acceptable_codes: Optional[List[int]] = None,
    **request_kwargs: Any,
) -> StructuredResponse:
    response = await _request_base(
        download_json_content=download_json_content,
        download_text_content=download_text_content,
        rate_limiter=rate_limiter,
        session=session,
        acceptable_codes=acceptable_codes,
        **request_kwargs,
    )
    structured_response = StructuredResponse(
        url=response.url,
        status_code=response.status_code,
        headers=response.headers,
        json_content=response.json_content,
        text_content=response.text_content,
        structured_content=model(**response.json_content),
    )
    if retry_based_on_content_callback and retry_based_on_content_callback(
        structured_response.structured_content
    ):
        raise ContentCallbackError("Structured content retry callback returned True")
    return structured_response


async def request_structured(
    model: Type[BaseModel],
    url: Union[str, URL],
    config: RequestConfig = RequestConfig(),
    payload: Optional[BaseModel] = None,
) -> StructuredResponse:
    """Make a request and structure the response.

    This function forces the download of the json content to be deserialized as a
    pydantic model.

    Args:
        model: pydantic model to be used to structure the response content.
        url: url to send request.
        config: request configuration.
        payload: payload to be sent in the request as a structured pydantic model.

    Returns:
        Structured response content deserialized as a pydantic model.

    """
    new_config = config.model_copy(update=dict(download_json_content=True))
    logger.debug(f"Request started, {new_config.method} method to {url}...")

    rc = new_config.retryer.retrying_config
    new_callable = _request_structured.retry_with(**rc)  # type: ignore

    async with aiohttp.ClientSession() as new_session:
        structured_response: StructuredResponse = await new_callable(
            model=model,
            retry_based_on_content_callback=(
                new_config.retryer.retry_based_on_content_callback
            ),
            download_json_content=new_config.download_json_content,
            download_text_content=new_config.download_text_content,
            rate_limiter=new_config.rate_limiter,
            session=new_config.session or new_session,
            acceptable_codes=new_config.acceptable_codes,
            **_build_kwargs(url=url, config=new_config, payload=payload),
        )

    logger.debug("Request finished!")
    return structured_response


async def request_structured_many(
    model: Type[BaseModel],
    urls: List[Union[str, URL]],
    config: RequestConfig = RequestConfig(),
    payloads: Optional[List[BaseModel]] = None,
) -> List[StructuredResponse]:
    """Make many requests and structure the responses.

    Args:
        model: pydantic model to be used to structure the response.
        urls: list of urls to send requests.
        config: request configuration.
        payloads: list of payloads as structured pydantic models.

    Returns:
        List of structured response content deserialized as a pydantic model.

    """
    payloads = _normalize_payloads(urls=urls, payloads=payloads)
    coroutines = (
        [
            request_structured(
                model=model,
                url=url,
                config=config,
                payload=payload,
            )
            for url, payload in zip(urls, payloads)
        ]
        if payloads
        else [
            request_structured(
                model=model,
                url=url,
                config=config,
            )
            for url in urls
        ]
    )

    results: List[StructuredResponse] = await asyncio.gather(*coroutines)
    return results


async def poll(
    model: Type[BaseModel],
    poll_config: PollConfig,
    url: Union[str, URL],
    config: RequestConfig = RequestConfig(),
) -> StructuredResponse:
    """Poll a url until a condition is met.

    Args:
        url: url to be polled.
        config: request configuration.
        model: pydantic model to be used to structure the response.
        poll_config: poll configuration.

    Returns:
        Structured response.

    """
    logger.debug(f"Polling {url}...")
    start_time = datetime.datetime.utcnow()
    elapsed_time = datetime.timedelta(seconds=0)
    while elapsed_time.total_seconds() < poll_config.timeout:
        response = await request_structured(model=model, url=url, config=config)
        if poll_config.success_condition(response.structured_content):
            logger.debug("Condition met, polling finished!")
            return response
        await asyncio.sleep(poll_config.interval)
        elapsed_time = datetime.datetime.utcnow() - start_time
        logger.debug(f"Condition not met yet. Elapsed time: {elapsed_time} seconds...")
    raise PollError("Timeout reached")



import asyncio
from functools import cached_property
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Type

import aiohttp
import tenacity
from aiohttp import ClientResponseError
from loguru import logger
from pydantic import BaseModel, ConfigDict, Field, JsonValue, computed_field
from pyrate_limiter import Duration, InMemoryBucket, Limiter, Rate
from yarl import URL

from biar.errors import ContentCallbackError, ResponseEvaluationError


class ProxyConfig(BaseModel):
    """Proxy configuration.

    Attributes:
        host: proxy address.
        headers: additional configuration required by the proxy.
        ssl_cadata: certificate as a string required by some proxies to use SSL.

    """

    host: str
    headers: Optional[Dict[str, Any]] = None
    ssl_cadata: Optional[str] = None


class Response(BaseModel):
    """Attributes from the http request response.

    Attributes:
        url: final url after (possible) redirects.
        status_code: HTTP status code.
        headers: headers in the response.
        json_content: response content as json dict.
        text_content: raw response content as a string.

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    url: URL
    status_code: int
    headers: Dict[str, Any] = Field(default_factory=dict)
    json_content: Dict[str, JsonValue] = Field(default_factory=dict)
    text_content: str = ""


class StructuredResponse(Response):
    """Attributes from the http request response.

    Attributes:
        url: final url after (possible) redirects.
        status_code: HTTP status code.
        headers: headers in the response.
        json_content: response content as json dict.
        text_content: raw response content as a string.
        structured_content: response content as a pydantic model.

    """

    structured_content: Any


class Retryer(BaseModel):
    """Retry logic with exponential backoff strategy.

    Attributes:
        attempts: number of attempts.
            `attempts=1` means only one try and no subsequent retry attempts.
        min_delay: number of seconds as the starting delay.
        max_delay: number of seconds as the maximum achieving delay.
        retry_if_exception_in: retry if exception found in this tuple.
            A ResponseEvaluationError is always added dynamically to be retried.

    """

    attempts: int = 1
    min_delay: int = 0
    max_delay: int = 10
    retry_if_exception_in: Tuple[Type[BaseException], ...] = (
        ClientResponseError,
        asyncio.TimeoutError,
        ContentCallbackError,
    )
    retry_based_on_content_callback: Optional[Callable[[BaseModel], bool]] = None

    @property
    def retrying_config(self) -> Dict[str, Any]:
        """Configuration for retrying logic.

        Changing arguments at run time reference:
        https://github.com/jd/tenacity#changing-arguments-at-run-time

        Returns:
            kwargs dictionary for tenacity.BaseRetrying.

        """
        return dict(
            stop=tenacity.stop_after_attempt(self.attempts),
            retry=tenacity.retry_if_exception_type(
                exception_types=self.retry_if_exception_in + (ResponseEvaluationError,)
            ),
            wait=tenacity.wait_exponential(min=self.min_delay, max=self.max_delay),
            reraise=True,
            before_sleep=tenacity.before_sleep_log(
                logger=logger, log_level="DEBUG"  # type: ignore[arg-type]
            ),
        )


class RateLimiter(BaseModel):
    """Limit the number of requests in a given time frame.

    Attributes:
        rate: number of requests allowed in the given time frame.
        time_frame: number of seconds for the time frame.
        identity: identification for the rate-limiting bucket.
            Same identity can be used universally for all endpoints in a given host, if
            the API have a global limit. If the API have different limits for each
            endpoint, different identities can be used as well.

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    rate: int = 10
    time_frame: int = 1
    identity: str = "default"

    @computed_field  # type: ignore[misc]
    @cached_property
    def limiter(self) -> Limiter:
        """In memory bucket to limit the number of requests."""
        return Limiter(
            InMemoryBucket(
                rates=[
                    Rate(
                        limit=self.rate,
                        interval=self.time_frame * Duration.SECOND.value,
                    )
                ]
            ),
            raise_when_fail=False,
            max_delay=Duration.MINUTE.value,
        )


class RequestConfig(BaseModel):
    """Base configuration for a request.

    Attributes:
        method: http method to be used.
        download_json_content: if true will await for json content download.
        download_text_content: if true will await for text content download.
        proxy_config: proxy configuration.
        rate_limiter: rate limiting configuration.
        retryer: retry logic configuration.
        timeout: maximum number of seconds for timeout.
            By default, is 300 seconds (5 minutes).
        use_random_user_agent: if true will use a random user agent.
        user_agent_list: list of user agents to be randomly selected.
            By default, it uses a sample from `biar.user_agents` module.
        bearer_token: bearer token to be used in the request.
        headers: headers dictionary to use in request.
        params: parameters dictionary to use in request.
        session: aiohttp session to be used in request.
            If the user wants to use a custom session and handle its lifecycle, it can
            be passed here.
        acceptable_codes: list of acceptable status codes.
            If the response status code is not in this list, an exception will be
            raised. By default, it only accepts 200.

    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    method: Literal["GET", "POST", "PUT", "DELETE"] = "GET"
    download_json_content: bool = True
    download_text_content: bool = True
    proxy_config: Optional[ProxyConfig] = None
    rate_limiter: RateLimiter = RateLimiter()
    retryer: Retryer = Retryer()
    timeout: int = 300
    use_random_user_agent: bool = True
    user_agent_list: Optional[List[str]] = None
    bearer_token: Optional[str] = None
    headers: Optional[Dict[str, str]] = None
    params: Optional[Dict[str, Any]] = None
    session: Optional[aiohttp.ClientSession] = None
    acceptable_codes: Optional[List[int]] = None


class PollConfig(BaseModel):
    """Poll configuration model.

    Args:
        timeout: timeout in seconds.
        interval: interval in seconds between polls.
        success_condition: callback to be called after each poll.
            The callback should return True if the polling should stop.

    """

    timeout: float = 60 * 5
    interval: float = 5
    success_condition: Callable[[BaseModel], bool]



import os
from typing import Dict

from setuptools import find_packages, setup

about: Dict[str, str] = {}
with open(
    file=os.path.join("biar", "__metadata__.py"),
    mode="r",
    encoding="utf-8",
) as f:
    exec(f.read(), about)

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    author=about["__author__"],
    license=about["__license__"],
    url=about["__url__"],
    packages=find_packages(
        exclude=["tests", "pipenv", "env", "venv", "htmlcov", ".pytest_cache", "pip"]
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.7, <4",
    install_requires=requirements,
)



import random
from typing import List, Optional

USER_AGENTS = [
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.80 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 12_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.80 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.80 Safari/537.36",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 15_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/98.0.4758.85 Mobile/15E148 Safari/604.1",
    "Mozilla/5.0 (Linux; Android 10) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.87 Mobile Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:96.0) Gecko/20100101 Firefox/96.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 12.2; rv:96.0) Gecko/20100101 Firefox/96.0",
    "Mozilla/5.0 (X11; Linux i686; rv:96.0) Gecko/20100101 Firefox/96.0",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 12_2 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) FxiOS/96.0 Mobile/15E148 Safari/605.1.15",
    "Mozilla/5.0 (Android 12; Mobile; rv:68.0) Gecko/68.0 Firefox/96.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 12_2) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/15.2 Safari/605.1.15",
    "Mozilla/5.0 (iPhone; CPU iPhone OS 15_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/15.2 Mobile/15E148 Safari/604.1",
]


def get_user_agent(user_agent_list: Optional[List[str]] = None) -> str:
    """Get a random User-Agent.

    The list needs to be updated from time to time from this database:
    https://www.whatismybrowser.com/guides/the-latest-user-agent/

    Returns:
        User-Agent string.

    """
    return random.choice(user_agent_list or USER_AGENTS)
